(* Time-stamp: <modified the 13/12/2012 (at 11:49) by Erwan Jahier> *)

(** Static evaluation of clocks. *)

open UnifyClock

(** [f lxm ids s ve cl] checks that [ve] is well-clocked (i.e., for node calls,
   it checks that the argument and the parameter clocks are compatible),
   and returns a clock profile that contains all the necessary information
   so that the caller can perform additional clock checks.
*)
val f : Lxm.t -> Lic.id_solver -> subst -> Lic.val_exp -> Lic.clock list -> 
  Lic.val_exp * Lic.id_clock list * subst


(** [check_res lxm cel cil] checks that the expected output clock
   profile of an expression "cil" is compatible with the result clocks
   "cel" of the expression.

   For instance, in order to clock check the equation 

      (x,y)=toto(a,b);

   it checks that the clock of "(x,y)" (cel) is compatible with the 
   output clock profile of node toto (cil).

*)

val check_res : Lxm.t -> subst -> Lic.left list -> Lic.id_clock list -> unit

