(** Time-stamp: <modified the 13/12/2012 (at 11:32) by Erwan Jahier> *)
 
  
open AstPredef
open AstV6
open AstCore
open Lic
open Printf
open Lxm
open Errors

(***
MESSAGES D'ERREUR :
Pour rester homogne, utiliser les fonctions de LicEvalType:
  raise_type_error (provided: string) (expect: string) (msg: string)
  raise_arity_error (msg:string) (provided:int) (expect:int)
*)
let raise_type_error = LicEvalType.raise_type_error
let raise_arity_error = LicEvalType.raise_arity_error
exception EvalType_error = LicEvalType.EvalType_error
exception EvalConst_error = LicEvalConst.EvalConst_error

let dbgpoly = Some (Verbose.get_flag "poly")

(******************************************************************************)
let finish_me msg = print_string ("\n\tXXX evalType.ml:"^msg^" ->  finish me!\n")


(********************************
ACCES AUX INFOS DEJA COMPILEES,
SOLUTION INTERMEDIARE  A REVOIR :
voir UglyStuff
*)


let rec (f : Lic.id_solver -> Lic.val_exp -> Lic.val_exp * Lic.type_ list) =
  fun id_solver ve ->
    let ve_core, tl =
      match ve.ve_core with
        | CallByPosLic ({it=posop; src=lxm}, OperLic args) -> (
            let posop_opt, args, tl = 
              try eval_by_pos_type id_solver posop lxm args
              with EvalType_error msg -> 
                raise (Compile_error(lxm, "type error: "^msg))
            in
            let ve = 
              match posop_opt with
                | None -> CallByPosLic ({it=posop; src=lxm}, OperLic args)
                | Some posop -> CallByPosLic ({it=posop; src=lxm}, OperLic args)
            in
              ve, tl
          )
        | CallByNameLic ({it=nmop; src=lxm}, nmargs ) ->
            let nmargs, tl = 
              try eval_by_name_type id_solver nmop lxm nmargs
              with EvalConst_error msg ->
                raise (Compile_error(lxm, "\n*** can't eval constant: "^msg))
            in
              CallByNameLic ({it=nmop; src=lxm}, nmargs ), tl
    in
      { ve_core = ve_core; ve_typ = tl ; ve_clk = ve.ve_clk }, tl
      

and eval_by_pos_type
   (id_solver: Lic.id_solver)
   (posop: Lic.by_pos_op)
   (lxm: Lxm.t)
   (args: Lic.val_exp list)
: (
      Lic.by_pos_op option (* For op that hold a val_exp, we return the modified op *)
      * Lic.val_exp list   (* The args with type info added *)
      * Lic.type_ list     (* The type of the val_exp "posop(args)" *)
  ) =
    match posop with
      | PREDEF_CALL (op,sargs) -> (
          let args, targs = List.split (List.map (f id_solver) args) in
          (* ICI pas de matches possible ? *)
          let tve = LicEvalType.f id_solver op lxm sargs targs in
            None, args, tve
        )
      | Lic.CALL nkf -> 
         let node_key = nkf.it in
         (* let node_exp_eff = id_solver.id2node node_key lxm in *)
         let node_exp_eff = UglyStuff.node_exp_of_node_key id_solver node_key lxm in
         let (lti, lto) = Lic.profile_of_node_exp node_exp_eff in
         let args, t_argsl = List.split (List.map (f id_solver) args) in
         let t_args = List.flatten t_argsl in
         let llti = List.length lti and lt_args = List.length t_args in
         let _ = if llti <> lt_args then
            raise_arity_error "" lt_args llti
         in
         (* lti = expecteds, t_args = given *)
         let tmatches = try UnifyType.is_matched lti t_args 
         with UnifyType.Match_failed msg -> (
            let msg' = Printf.sprintf 
               "\n***    while unifing (%s) with (%s)"
               (Lic.string_of_type_list lti)
               (Lic.string_of_type_list t_args)
            in raise (EvalType_error(msg'^msg))
         ) in
         let tve = match tmatches with
            | [] -> lto
            | _  -> 
               Verbose.exe ~flag:dbgpoly (fun () ->
                  Printf.fprintf stderr "#DBG: EvalType of CALL '%s' (%s)\n"
                     (Lic.string_of_node_key node_key) 
                     (Lxm.details lxm);
                  Printf.fprintf stderr "#     unifying '%s' with '%s'\n"
                     (Lic.string_of_type_list lti)
                     (Lic.string_of_type_list t_args) ;
                  Printf.fprintf stderr "#     required matches %s\n"
                     (Lic.string_of_type_matches tmatches) 
               );
               Lic.apply_type_matches tmatches lto
         in
          (None, args, tve)
      | Lic.CONST_REF idl ->
         let ceff = UglyStuff.const_eff_of_item_key id_solver idl lxm in
         let tve = [Lic.type_of_const ceff] in
         None, [], tve
      | Lic.VAR_REF id ->
         let tve = [
            (UglyStuff.var_info_of_ident id_solver id lxm).var_type_eff
         ] in
         None, [], tve
      | Lic.WITH(ve) ->
          let ve, tve = f id_solver ve in
            Some(Lic.WITH(ve)), [], tve
            
      | Lic.TUPLE ->
          let args, targs = List.split (List.map (f id_solver) args) in
            None, args, List.flatten targs

      | Lic.CONCAT ->
         let args, targs = List.split (List.map (f id_solver) args) in
         let tve = match targs with
         | [ [Array_type_eff (t0, s0)]; [Array_type_eff (t1, s1)]] ->
            if t0 = t1 then [Array_type_eff (t0, s0+s1)]
            else
               raise_type_error (List.flatten targs) []
                  "two arrays of the same type was expected"
         | _ -> 
            raise_arity_error "" (List.length args) 2
          in
            None, args, tve
      | Lic.STRUCT_ACCESS (fid) ->
          assert (List.length args = 1);          
          let arg, targ = f id_solver (List.hd args) in
          let teff_field =
            match targ with
              | [Struct_type_eff (name, fl)] -> (
                  try fst (List.assoc fid fl)
                  with Not_found ->
                    raise (
                      EvalType_error
                        (Printf.sprintf "%s is not a field of struct %s"
                           (Ident.to_string fid)
                           (Lic.string_of_type (List.hd targ))))
                )
              | [x] -> raise_type_error [x] [] "some struct type was expected"
              |  x -> raise_arity_error "" (List.length x) 1 
          in
            None, [arg], [teff_field]


      | Lic.ARRAY_ACCES(i) ->
          assert (List.length args = 1);          
          let arg, targ = f id_solver (List.hd args) in
          let sz, teff =
            match targ with
              | [Array_type_eff(teff_elt, size)] -> size, teff_elt
              | _ ->
                  let msg = 
                    "\n*** Type error: '"^(Lic.string_of_type_list targ) ^
                      "' was expected to be an array"
                  in
                    raise (Compile_error(lxm, msg))
          in
          let _ = if ((i >= 0) && (i < sz)) then () else
            raise(EvalType_error(sprintf "array index %d out of bounds 0..%d" i (sz-1)))
          in
            None, [arg], [teff]

      | Lic.ARRAY_SLICE(sieff) ->
          assert (List.length args = 1);          
          let arg, targ = f id_solver (List.hd args) in
          let sz, teff_elt =
            match targ with
              | [Array_type_eff(teff_elt, size)] -> size, teff_elt
              | _ ->
                  raise (Compile_error(
                           lxm, "\n*** Type error: '" ^ 
                             (Lic.string_of_type_list targ) ^ 
                             "' was expected to be an array"))
          in
          let _ = if ((sieff.se_first >= 0) && (sieff.se_first < sz)) then () else
            raise(
              EvalType_error(sprintf "array index %d out of bounds 0..%d" 
                               sieff.se_first (sz-1)))
          in
          let _ = if ((sieff.se_last >= 0) && (sieff.se_last < sz)) then () else
            raise(
              EvalType_error(sprintf "array index %d out of bounds 0..%d" 
                               sieff.se_last (sz-1)))
          in
            None, [arg], [Array_type_eff(teff_elt, sieff.se_width)]

      | Lic.HAT(size,ceff) -> 
          let ceff, teff_list = f id_solver ceff in
          let tl = List.map (fun teff -> Array_type_eff(teff, size)) teff_list in
            Some(Lic.HAT(size,ceff)), [], tl
              
      | Lic.ARRAY(args) ->
          (* check that args are of the same type *)
          let args, targs = List.split (List.map (f id_solver) args) in
          let teff_elt =
            List.fold_left
              (fun acc teff ->
                 match acc with
                   | [] -> teff
                   | [sacc] -> if acc = teff then acc else
                       raise(EvalType_error(
                               "all array elements should be of the same type"))
                   | _ -> assert false
              )
              []
              targs
          in
          let tve =
            assert (List.length teff_elt = 1);
            [Array_type_eff(List.hd teff_elt, List.length args)]
          in
            Some(Lic.ARRAY(args)), [], tve

      | Lic.WHEN clk_exp -> (
          let args, targs = List.split (List.map (f id_solver) args) in
          let _ = match clk_exp with
            | Base -> ()
            | NamedClock( { it = (cc,cv) ; src = lxm }) -> 
                let vi = id_solver.id2var (Ident.to_idref cv) lxm in
                  (match vi.var_type_eff with
	             | Lic.Bool_type_eff
                     | Lic.Enum_type_eff _ -> ()
	             | teff  -> 
		         let msg = "the type of a clock cannot be " ^
                           (Lic.string_of_type teff) 
		         in
		           raise(Compile_error(lxm,msg))
                  )
          in
            match targs with
              | [teff] -> None, args, teff
              | _ -> raise_arity_error "" (List.length targs) 1 
        )
      | Lic.ARROW
      | Lic.FBY -> (
          let args, targs = List.split (List.map (f id_solver) args) in
            match targs with
              | [init; teff] -> if init = teff then None, args, teff else 
                  raise(EvalType_error("type mismatch. "))
              | _ -> raise_arity_error "" (List.length targs) 2
        )
      | Lic.CURRENT 
      | Lic.PRE -> (
          let args, targs = List.split (List.map (f id_solver) args) in
            match targs with
              | [teff] -> None, args, teff
              | _ -> raise_arity_error "" (List.length targs) 1
        )
      | Lic.MERGE _ -> finish_me "merge"; assert false


(**
   Juste pour les structures ...
*)
and eval_by_name_type
   (id_solver:  Lic.id_solver)
   (namop: Lic.by_name_op)
   (lxm: Lxm.t)
   (namargs: (Ident.t Lxm.srcflagged * Lic.val_exp) list )
(* renvoie la lsite de modif de champs compile + le type du rsultat *)
:  (Ident.t Lxm.srcflagged * Lic.val_exp) list * Lic.type_ list
= match namop with

   | Lic.STRUCT_anonymous -> 
      (* ??? comment faire ici pour recuperer son type ???
             il faut que je recherche  l'aide des noms de champs
             le type structure qui va bien !

             - creer une table [liste des noms de champs -> ident de type structure] ?
             - rajouter dans la table a sa creation une entree dont le nom
             est compos du nom des champs ?
      *)
      finish_me "anonymous struct not yet supported"; 
      assert false
      
   | Lic.STRUCT (pn,opid) -> 
      let struct_type = id_solver.id2type opid lxm in
      match struct_type with
      | Struct_type_eff(sn, fl) -> 
         let do_field_assign (fn, fv) =
         (* traitement d'un ''field_name  = field_value'' *)
            let (ft,fopt) = try
               List.assoc fn.it fl
               with Not_found -> 
                  let msg = "type error: bad field"^(Ident.to_string fn.it) in
                  raise (Compile_error(lxm, msg))
            in
            (* let's check the type of fv *)
            let fv, fv_type = f id_solver fv in
            if fv_type = [ft] then (fn,fv)
            else raise_type_error fv_type [ft]
                ("while checking struct field "^(Lxm.details fn.src))
         in
         let namargs = List.map do_field_assign namargs
         in (namargs, [struct_type])
      | _ -> raise (Compile_error(lxm, "type error: a structure is expected"))
