(* Time-stamp: <modified the 13/12/2012 (at 16:16) by Erwan Jahier> *)


(** Define the Data Structure representing Compiled programs. *)

(**
        Dfinition des structures de donnes utilise pour la compil,
        plus des utilitaires pour les messages d'erreurs, de bug etc.
        N.B. on utilise beaucoup l'adjectif "effectif", qui signifie
        simplement "correct".

        REMARQUE GENERALE :

        D'une manire gnrale, la compil d'une entit syntaxique
        "toto" est implmente par une fonction check_toto, qui
        prend en entre (entr'autre) un toto et renvoie un
        toto.

        TYPES DE DONNEES :

        - type_ :
        dnotation de type effectif, implmente l'quivalence des types,
        construit  partir d'une type_exp.

        - const :
        dnotation de constante effective,
        construit  partir d'une val_exp => IL S'AGIT DE LA REPRESENTATION
        INTERNE DES CONSTANTES STATIQUES 

        - var_info :
        dclaration de variable,
        construit  partir de var_info.

        - val :
        union entre const et var_info.

        - slice_info :
        dnotation de tranche de tableau,
        construit  partir de slice_info.

        - left :
        version compile de left_part

        - eq_info :
        version compile de eq_info 

        - node_exp :
        dnotation d'opration,
        peut tre  predef ou utilisateur,
        construit  partir de node_exp.

        - static_arg :
        dclaration d'un static arg

        - pack_env :
        la "grosse" structure de donnes qui gre la compilation
        des packages => implmente dans CheckGlobal pour la partie type/const/function
        (initialisation) et dans CheckNode pour la partie node/template qui
        est faite  la demande.

        - local_env :
        structure qui gre l'environnement de compilation
        d'un noeud/template.

        TYPES FONCTIONNEL :

        - id_solver (en fait, une structure qui contient plusieurs fonctions,
          une pour traiter les constantes, une pour traiter les types)

        UTILITAIRES :

        - type_of_const : renvoie le type d'une const
        - string_of_type : pretty-print d'un type 
        - string_of_const : pretty-print d'une const 
        - string_of_node_key : pretty-print d'un node_key
        _ string_of_slice :

----------------------------------------------------------------------*)

open Errors 
open Printf 
open Lxm
open AstV6
(* open AstCore *)

let dbg = Some (Verbose.get_flag "lazyc")

(*---------------------------------------------------------------------
Type : id_solver
-----------------------------------------------------------------------
        Joue le rle d'environnemnt : contient des fonctions
        pour rsoudre les rferences aux idents.
        (voir par exemple EvalConst, EvalType)

N.B.

- On fournit les constructeurs des id_solver courants

----------------------------------------------------------------------*)
type id_solver = {
  (* XXX I should not have [idref] in this module !!! *)
  id2const : Ident.idref -> Lxm.t -> const;
  id2type  : Ident.idref -> Lxm.t -> type_;
  id2node  : Ident.idref -> static_arg list -> Lxm.t -> node_exp;
  id2var   : Ident.idref -> Lxm.t -> var_info;
  global_symbols  : AstTabSymbol.t;
}

(*---------------------------------------------------------------------
  Type : type
  -----------------------------------------------------------------------
  Dnotation de type immdiat : l'quivalence smantique des types
  EST l'quivalence structurelle des types. 
  Par rapport  une type_exp :
  - pas d'alias
  - taille des tableaux rsolues
  ----------------------------------------------------------------------*)

and type_ =
  | Bool_type_eff
  | Int_type_eff
  | Real_type_eff
  | External_type_eff of Ident.long
  | Abstract_type_eff of Ident.long * type_ 
  | Enum_type_eff     of Ident.long * (Ident.long list)
  | Array_type_eff    of type_ * int
  | Struct_type_eff   of 
      Ident.long * (Ident.t * (type_ * const option)) list
  | TypeVar of type_var
      (* [Overload] is like [Any], except that it can only be [int] or [real] *)
and type_var =
  | Any
  | AnyNum

(* Utile :
- arguments et profils
*)
(* A VIRER A MOYEN TERME ! *)
and node_profile = (Ident.t * type_) list * (Ident.t * type_) list
and profile = type_ list * type_ list

and slice_info = {
(** Dnotation de tranche de tableau correcte :
    si A est le tableau d'entre, alors S est le tableau
    de sortie avec :
    S[i] = A[first + i*step] pour i = 0 .. width
*)
  se_first : int;
  se_last : int;
  se_step : int;
  se_width : int; (* -> size? *)
}

and left =
(** Version checke des left_part
  (les idents, les index et les tranches sont rsolus)

  N.B. On conserve aussi le type effectif de chaque noeud
  bien qu'il soit possible de le retrouver. 

  N.B. On garde aussi l'info source des idents au cas ou.*)
  | LeftVarLic  of (var_info * Lxm.t)
  | LeftFieldLic of (left * Ident.t * type_)
  | LeftArrayLic of (left * int * type_)
  | LeftSliceLic of (left * slice_info * type_)


and eq_info = left list * val_exp

and val_exp =
    { ve_core :  val_exp_core ; 
      ve_typ : type_ list ;
        (* An empty list means that its type has not been computed (EvalType.f) yet.
           a cleaner solution would be to define two versions of val_exp: one with
           type info, and one without. But it is a big mutually recursive thing,
           and doing that would be a little bit heavy...
        *)
      ve_clk : clock list
        (* ditto *)
    }
(** CallByPosLicest (sans doute ?)
   le BON endroit pour stocker l'information de 'matches',
   i.e. est-ce qu'un 'type_matches' a t ncessaire
   pour typer l'appel de l'oprateur ?
*)
and val_exp_core =
  | CallByPosLic of (by_pos_op srcflagged * operands)
  | CallByNameLic of
      (by_name_op srcflagged * (Ident.t srcflagged * val_exp) list)

and operands = OperLic of val_exp list

and by_name_op =
  | STRUCT of Ident.pack_name * Ident.idref
  | STRUCT_anonymous


and by_pos_op =
  | PREDEF_CALL of AstPredef.op * static_arg list
  | CALL of node_key srcflagged
  (* | IDENT of Ident.idref (* should be an Ident.t or long, really... *) *)
  | CONST_REF of Ident.long
  | VAR_REF of Ident.t

  | PRE
  | ARROW
  | FBY
  | CURRENT

  | WHEN of AstCore.clock_exp
  | TUPLE
  | WITH of val_exp

  | CONCAT
  | HAT of int * val_exp
  | ARRAY of val_exp  list
  | STRUCT_ACCESS of Ident.t

(* those are different from [by_pos_op] *)

  | ARRAY_ACCES of int
  | ARRAY_SLICE of slice_info
  | MERGE       of (Ident.t * (Ident.t list))  


(*---------------------------------------------------------------------
  Type : const
  -----------------------------------------------------------------------
  Dnotation de constante immdiate
  N.B. les const "portent" leur type :
  - il est implicite pour bool, int, real, 
  - explicite pour extern, enum et struct
  - pour array c'est le TYPE DES ELEMENTS QU'ON TRIMBALE 
  VOIR => type_of_const
  ----------------------------------------------------------------------*)
and const =
    (* type predef *)
    Bool_const_eff of bool
  | Int_const_eff of int (* XXX should be a string ! *)
  | Real_const_eff of string
      (* type atomique non predef : on prcise le type *)
  | Extern_const_eff of (Ident.long * type_)
  | Abstract_const_eff of (Ident.long * type_ * const * bool)
      (* if the abstract const is exported (i.e., defined in the provided part),
         then the bool flag is set to true. Well, it is not really an abstract
	 constant in that case...

         nb: well, are exported abstract constant useful at all?  I don't
         know, but it is not a reason to handle them unproperly...
      *)
  | Enum_const_eff   of (Ident.long * type_)
      (* type_ structure : liste (champ,valeur) + type_ structure *)
  | Struct_const_eff of ((Ident.t * const) list * type_)
      (* type_ tableau : liste des valeurs + type_ des elts + taille *)
  | Array_const_eff of (const list * type_)
  | Tuple_const_eff of const list
      (*---------------------------------------------------------------------
        Type: val   
        -----------------------------------------------------------------------
        Une constante ou une variable
        => item de la table des symboles de valeurs
        ----------------------------------------------------------------------*)
(* and val = *)
(*     ConstLic of const *)
(*   | VarLic of var_info *)
      (*---------------------------------------------------------------------
        Type: var_info      
        -----------------------------------------------------------------------
        Info associe  un ident de variable
        ----------------------------------------------------------------------*)
      (* ICI  completer/modifier sans doute *)
and var_info = {
  var_name_eff   : Ident.t;
  var_nature_eff : AstCore.var_nature;
  var_number_eff : int;
  var_type_eff   : type_;
  var_clock_eff  : id_clock;
}
and id_clock = Ident.t * clock
      (* 
         A pair made of an ident and its clock.

         The ident is used to relate arguments and parameters
         when clocking node calls (in order to be able to transmit
         the clocking constraints between the node I/O, in EvalClock.f).
      *)
and clock =
  | BaseLic
  | ClockVar of int (* to deal with polymorphic clocks (i.e., constants) *)
  | On of Ident.clk * clock 
      (* The clock expression, and the effective clock the clock var
         e.g., in 
            On((clk_constructor, clk_var), clk )
         clk is the clock of clk_var
      *)

(**********************************************************************************)
(** [node_exp] correspond  une instance de template (ou, cas
    limite, de noeud sans param statique).

    La cl est un couple ident/liste d'arguments statiques effectifs

    N.B. une horloge formelle est soit None (base) soit l'index d'une
    entre (0..nb entres-1). Les formal-clocks sont cres au cours du
    type-checking (et pas du clock-checking)

*) 
and node_exp = {
  node_key_eff : node_key;
  inlist_eff   : var_info list;
  outlist_eff  : var_info list;
  loclist_eff  : var_info list option; (* None => extern or abstract *)
  def_eff      : node_def;
  has_mem_eff  : bool;
  is_safe_eff  : bool;
  (* is_polym_eff : bool *)
}

and type_matches = (type_var * type_) list

and node_def =
  | ExternLic
  | MetaOpLic of node_key
   (* ICI A QUOI CA SERT ???? *)
  | AbstractLic of node_exp option (* None if extern in the provide part *)
  | BodyLic of node_body

and node_body = {
  asserts_eff : (val_exp srcflagged) list;
  eqs_eff     : (eq_info srcflagged) list;
}

(* key used  for type, constant, and clock tables *)
and item_key = Ident.long
and node_key = item_key * static_arg list
and static_arg =
	(* may be a tuple *)
  | ConstStaticArgLic of (Ident.t * const)
  | TypeStaticArgLic of (Ident.t * type_)
  (* | NodeStaticArgLic of (Ident.t * sarg_node_eff * node_exp) *)
  | NodeStaticArgLic of (Ident.t * node_key)

and sarg_node_eff = node_key * var_info list * var_info list

(****************************************************************************)
(** Type check_flag

   Au cours du check, on conserve le statut des idents :
   
   - Checking => en cours de traitement, permet de lever les rcursions
   - Checked  => trait et correct
   - Incorrect => dj marqu comme incorrect (pas besoin d'un nouveau
   message d'erreur)
*)
type 'a check_flag =
    Checking
  | Checked of 'a
  | Incorrect


let (profile_of_node_exp : node_exp -> profile) =
  fun ne -> 
    List.map (fun vi -> vi.var_type_eff) ne.inlist_eff,
    List.map (fun vi -> vi.var_type_eff) ne.outlist_eff
    
(****************************************************************************)
(* currently not used *)


(* type world_env = { *)
(*   wenv_src : AstV6.pack_or_model list; *)
(*   wenv_mod_srcs : (Ident.t, AstV6.model_info srcflagged) Hashtbl.t ; *)
(*   wenv_pack_srcs :  (Ident.t, AstV6.pack_info srcflagged) Hashtbl.t ; *)
(*   wenv_pack_envs : (Ident.t, pack_env) Hashtbl.t ; *)
(* } *)
(* and pack_env = { *)
(*   penv_world : world_env ; *)
(*   (* penv_src : AstV6.package ; *) *)
(*   penv_type_table  : (Ident.t, type check_flag)   Hashtbl.t ; *)
(*   penv_const_table : (Ident.t, const check_flag)  Hashtbl.t ; *)
(*   penv_oper_table  : (Ident.t, node_half) Hashtbl.t ; *)
(*   penv_node_table : (node_key, node_exp check_flag) Hashtbl.t *)
(* } *)

(* the local tables are indexed by Ident.t, because local idents (var,const, flow)
   cannot have any package name. 

   and for nodes, the only possibility to have an entry in this table is via the
   static parameters.
   
   i.e.  
        min_4 = min_n<< 4, toto<<2>> >> ;

   is not allowed (I think). One has to write something like :

      toto_2 = toto<<2>>;
      min_4 = min_n<< 4, toto_2 >> ;

   It would not be difficult to handle that here though.
*)
type local_env = {
  lenv_node_key : node_key ;
(*   lenv_globals : pack_env ; *)
  lenv_types : (Ident.t, type_) Hashtbl.t ;
  lenv_const : (Ident.t, const) Hashtbl.t ; 
  (* lenv_nodes : (Ident.t, sarg_node_eff) Hashtbl.t ;  *)
  lenv_nodes : (Ident.t, node_key) Hashtbl.t ; 

  lenv_vars  : (Ident.t, var_info) Hashtbl.t ; 
}

(* Just to group those 2 ones *)
type node_env = {
  local : local_env;
  global: id_solver;
}


let (lookup_type: local_env -> Ident.idref -> Lxm.t -> type_) = 
  fun env id lxm -> 
    Hashtbl.find env.lenv_types (Ident.name_of_idref id)

let lookup_node
   (env: local_env)
   (id: Ident.idref)
   (sargs: static_arg list)
   (lxm: Lxm.t)
(* : sarg_node_eff =  *)
: node_key = 

    Hashtbl.find env.lenv_nodes (Ident.name_of_idref id)
    
let (lookup_const: local_env -> Ident.idref -> Lxm.t -> const) = 
  fun env id lmx ->  
    Hashtbl.find env.lenv_const (Ident.name_of_idref id) 
          
let (lookup_var: local_env -> Ident.t -> Lxm.t -> var_info) = 
  fun env id lmx ->  
    Hashtbl.find env.lenv_vars id


(****************************************************************************)
(** [type_are_compatible t1 t2] checks that t1 is compatible with t2, i.e., 
    if t1 = t2 or t1 is abstract and not t2.
*)
let (type_are_compatible : type_ -> type_ -> bool) =
  fun te1 te2 -> match te1, te2 with
    | External_type_eff (id1), External_type_eff (id2) -> id1 = id2
    | External_type_eff _, _ -> true
    | Abstract_type_eff _, _ -> true
    | TypeVar Any, _ -> true
    | _, TypeVar Any -> true
    | (TypeVar AnyNum), Real_type_eff | Real_type_eff, (TypeVar AnyNum)
    | (TypeVar AnyNum), Int_type_eff | Int_type_eff, (TypeVar AnyNum) -> true
    | t1, t2 -> t1 = t2

let rec (clock_are_equals : clock -> clock -> bool) =
  fun c1 c2 -> match c1, c2 with
    | On(cid1,_), On(cid2,_) -> cid1 = cid2
(* equivalent ? try both before commit !!! *)
(*     | On(_,c1), On(_,c2) -> clock_are_equals c1 c2 *)
    | c1, c2 -> c1 = c2

let (var_are_compatible : var_info -> var_info -> bool) =
  fun v1 v2 -> 
    (type_are_compatible v1.var_type_eff v2.var_type_eff) &&
      (clock_are_equals (snd v1.var_clock_eff) (snd v2.var_clock_eff))
  
let ident_of_type = function
  | Bool_type_eff -> Ident.out_of_pack "bool"
  | Int_type_eff -> Ident.out_of_pack "int"
  | Real_type_eff -> Ident.out_of_pack "real"
  | External_type_eff id
  | Abstract_type_eff (id, _)
  | Enum_type_eff     (id, _)
  | Struct_type_eff   (id, _) -> id
  | TypeVar Any -> Ident.out_of_pack "any"
  | (TypeVar AnyNum) -> Ident.out_of_pack "anynum"
  | _ -> assert false


(****************************************************************************)
(* Utilitaires lis aux node_key *)
let (make_simple_node_key : Ident.long -> node_key) =
  fun nkey -> (nkey, [])

(* OBSOLETE ET UN PEU FAUX ! *)
let rec (subst_type : type_ -> type_ -> type_) =
  fun t teff_ext -> match teff_ext with
      (* substitutes [t] in [teff_ext] *)
    | Bool_type_eff -> Bool_type_eff
    | Int_type_eff -> Int_type_eff
    | Real_type_eff -> Real_type_eff
    | External_type_eff(l) -> External_type_eff(l)
    | Abstract_type_eff(l,t) -> Abstract_type_eff(l,t)
    | Enum_type_eff(l,el) ->  Enum_type_eff(l,el)
    | Array_type_eff(teff_ext,i) -> 
        Array_type_eff(subst_type t teff_ext, i)
    | Struct_type_eff(l, fl) -> 
        Struct_type_eff(
          l, List.map (fun (id,(teff,copt)) -> (id,(subst_type t teff,copt))) fl)
    | TypeVar Any 
    | (TypeVar AnyNum) -> t

(* *) 
let rec subst_matches (matches: type_matches) (t: type_) : type_ =
   match t with
   | Bool_type_eff
   | Int_type_eff
   | Real_type_eff
   | External_type_eff _
   | Enum_type_eff _ -> t
   (* normallement, seul cas rcursif ? *)
   | Array_type_eff(telts,i) -> 
      Array_type_eff(subst_matches matches telts, i)
   (* NE DEVRAIENT PAS ETRE RECURSIFS
      on utilse paranoid au cas o ...
   *)
   | Abstract_type_eff(l,td) ->
      Verbose.exe ~flag:Global.paranoid ( fun () ->
         let t' = Abstract_type_eff(l,subst_matches matches td) in
         if t <> t' then
            assert false   
      );
      t
   | Struct_type_eff(l,fl) -> 
      Verbose.exe ~flag:Global.paranoid ( fun () ->
         let t' = Struct_type_eff(
            l, List.map (fun (id,(teff,copt)) -> (id,(subst_matches matches teff, copt))) fl)
         in
         if t <> t' then
            assert false   
      );
      t
   | TypeVar tvar ->
      try (List.assoc tvar matches) with Not_found -> t

let apply_type_matches (matches: type_matches) (tl: type_ list) : type_ list =
   match matches with
   | [] -> tl
   | _ -> List.map (subst_matches matches) tl


let rec (type_is_poly : type_ -> bool) =
  fun t -> match t with
    | Bool_type_eff
    | Int_type_eff 
    | Real_type_eff
    | External_type_eff _
    | Enum_type_eff(_)  -> false
    (* peut-tre un alias ! *)
    | Abstract_type_eff (id,te) -> type_is_poly te
    | TypeVar Any 
    | (TypeVar AnyNum) -> true
    | Array_type_eff(teff_ext,i) -> type_is_poly teff_ext
    | Struct_type_eff(l, fl) -> 
        List.exists (fun (_,(teff,_)) -> type_is_poly teff) fl

let node_is_poly (ne:node_exp) : bool =
   (* let it, ot = profile_of_node_exp ne in *)
   let varispoly v = type_is_poly v.var_type_eff in 
   List.exists varispoly ne.inlist_eff
   ||
   List.exists varispoly ne.outlist_eff

let node_is_extern (ne:node_exp) : bool =
   match ne.def_eff with
   | ExternLic -> true
   | _ -> false

(* Ne doit tre appele que pour les constantes simple
*)
let (type_of_const: const -> type_) =
  function
    | Bool_const_eff _ -> Bool_type_eff
    | Int_const_eff  _ -> Int_type_eff
    | Real_const_eff _ -> Real_type_eff
    | Extern_const_eff (s,  teff) -> teff
    | Abstract_const_eff (s,  teff, _v, _is_exported) -> teff
    | Enum_const_eff   (s,  teff) -> teff
    | Struct_const_eff (fl, teff) -> teff
    | Array_const_eff  (ct, teff) -> Array_type_eff (teff, List.length ct)
    | Tuple_const_eff cl -> 
		print_internal_error "Lic.type_of_const" "should not have been called for a tuple";
		assert false

let type_of_val_exp ve = ve.ve_typ

(* accepte un UNIQUE niveau de tuple
*)
let (types_of_const: const -> type_ list) =
	function
	| Tuple_const_eff cl -> List.map type_of_const cl
	| c -> [type_of_const c]

(* const list *)

(* Ignore the abstraction layer (necessary when expanding struct) *)
(* XXX not used anymore. This is very suspect... *)
let (true_type_of_const: const -> type_) =
  function
    | Abstract_const_eff (s,  teff, _v, _is_exported) -> teff
    | teff -> type_of_const teff


let (type_of_left: left -> type_) =
  function
    | LeftVarLic(vi,lxm)   -> vi.var_type_eff 
    | LeftFieldLic(_, _, t) -> t 
    | LeftArrayLic(_, _, t) -> t
    | LeftSliceLic(_, _, t) -> t 

 
let rec (var_info_of_left: left -> var_info) =
  function
    | LeftVarLic  (v, _) -> v
    | LeftFieldLic(left,_,_) -> var_info_of_left left
    | LeftArrayLic(left,_,_) -> var_info_of_left left 
    | LeftSliceLic(left,_,_) -> var_info_of_left left

let (clock_of_left: left -> clock) =
  fun left -> 
    snd (var_info_of_left left).var_clock_eff

(* utils N.B peut etre different de LicDump ! *)
let string_of_ident x =
   if !Global.no_prefix then
      Ident.no_pack_string_of_long x
   else
      Ident.string_of_long x

let rec string_of_type = function
  | Bool_type_eff -> "bool"
  | Int_type_eff  -> "int"
  | Real_type_eff -> "real"
  | External_type_eff (name) -> (string_of_ident name)
  | Abstract_type_eff (name, t) -> (string_of_ident name)
  | Enum_type_eff (name, _) -> (string_of_ident name)
  | Array_type_eff (ty, sz) ->
      Printf.sprintf "%s^%d" (string_of_type ty) sz
  | Struct_type_eff (name, _) -> (string_of_ident name)
  | TypeVar Any -> "any"
  | (TypeVar AnyNum) -> "anynum"

and string_of_type_list = function
  | []  -> ""
  | [x] -> string_of_type x
  | l   -> String.concat " * " (List.map string_of_type l)

and string_of_type_profile (i, o) =
  (string_of_type_list i)^" -> "^(string_of_type_list o)

and string_of_clock = function
  | BaseLic -> " on base"
  | ClockVar i -> " on 'c"^(string_of_int i)
  | On (id, ck) -> " on "^(Ident.string_of_clk id)^(string_of_clock ck)

and string_of_const = function
   | Bool_const_eff true -> "true"
   | Bool_const_eff false -> "false"
   | Int_const_eff i -> (sprintf "%d" i)
   | Real_const_eff r -> r
   | Extern_const_eff (s,_) -> (string_of_ident s)
   | Abstract_const_eff (s,t,v,_) -> (string_of_ident s)
   | Enum_const_eff   (s,_) -> (string_of_ident s)
   | Struct_const_eff (fl, t) -> 
      let string_of_field (id, veff) =
        (Ident.to_string id)^" = "^ (string_of_const veff)
      in
      Printf.sprintf "%s{%s}"
         (string_of_type t)
         (String.concat "; " (List.map string_of_field fl))
   | Array_const_eff (ctab, t) ->
      Printf.sprintf "[%s]"
         (String.concat ", " (List.map string_of_const ctab))
   | Tuple_const_eff   cl ->
      Printf.sprintf "(%s)"
         (String.concat ", " (List.map string_of_const cl))

and string_of_var_info x =
   (Ident.to_string x.var_name_eff) ^ ":"^(string_of_type x.var_type_eff)^(string_of_clock (snd x.var_clock_eff))
and string_of_var_list vl = String.concat " ; " (List.map string_of_var_info vl)

and string_of_node_key = function
| (ik, []) ->
   (string_of_ident ik)
| (ik, sargs)  -> Printf.sprintf "%s<<%s>>"
   (string_of_ident ik)
   (String.concat ", " (List.map string_of_static_arg sargs))

and string_of_static_arg = function
| ConstStaticArgLic(id, ceff) -> Printf.sprintf "const %s = %s" id (string_of_const ceff)
| TypeStaticArgLic (id, teff) -> Printf.sprintf "type %s = %s" id (string_of_type teff)
(* | NodeStaticArgLic (id, ((long,sargs), _, _), _) -> *)
| NodeStaticArgLic (id, nk) ->
   Printf.sprintf "node %s = %s" id (string_of_node_key nk)

and string_of_type_var tv = string_of_type (TypeVar tv)
and string_of_type_matches pm =
   let sotm (tv,t) = Printf.sprintf "%s <- %s"
      (string_of_type_var tv) (string_of_type t)
   in
   String.concat ", " (List.map sotm pm)

let string_of_node_exp ne =
  (Printf.sprintf "   node_key_eff = %s\n" (string_of_node_key ne.node_key_eff))
^ (Printf.sprintf "   inlist_eff   = %s\n" (string_of_var_list ne.inlist_eff))
^ (Printf.sprintf "   outlist_eff  = %s\n" (string_of_var_list ne.outlist_eff))
  (* ne.loclist_eff  : var_info list option; (* None => extern or abstract *) *)
  (* ne.def_eff      : node_def; *)
  (* ne.has_mem_eff  : bool; *)
  (* ne.is_safe_eff  : bool; *)

(* NodeStaticArgLic of (Ident.t * sarg_node_eff * node_exp) *)
(* sarg_node_eff = node_key * var_info list * var_info list *)

(* utile : liste standard de var_info a partir de liste de type *)
let create_var_list nat tl = (
   let pfx = match nat with
   | AstCore.VarInput -> "i"
   | AstCore.VarOutput -> "o"
   | AstCore.VarLocal -> assert false
   in
   let cpt = ref 0 in
   let dovar t = (
      let i = !cpt in
      let id = Printf.sprintf "%s%d" pfx i in
      incr cpt;
      {
         var_name_eff   = id;
         var_nature_eff = nat;
         var_number_eff = i;
         var_type_eff   = t;
         (* ???? *)
         var_clock_eff  = (id, BaseLic);
      }
   ) in List.map dovar tl 
)

(*---------------------------------------------------------------------
Une erreur associe  un noeud + 1 lexeme dans le fichier source
----------------------------------------------------------------------*)
exception Compile_node_error of node_key * Lxm.t * string
exception Global_node_error of node_key * string

let (make_local_env : node_key -> local_env) =
  fun nk ->
    Verbose.printf ~flag:dbg "#make_local_env %s\n" (string_of_node_key nk);
    let res =
      {
        lenv_node_key = nk;
        lenv_types = Hashtbl.create 0;
        lenv_const = Hashtbl.create 0;
        lenv_nodes = Hashtbl.create 0;
        lenv_vars  = Hashtbl.create 0;
      }
    in
      (* fill tables using static arg info *)
      List.iter
        (function
           | ConstStaticArgLic(id,ce) -> Hashtbl.add res.lenv_const id ce
           | TypeStaticArgLic(id,te)  -> Hashtbl.add res.lenv_types id te
           (* | NodeStaticArgLic(id, ne, _) -> Hashtbl.add res.lenv_nodes id ne *)
           | NodeStaticArgLic(id, nk) -> Hashtbl.add res.lenv_nodes id nk
        )
        (snd nk);
      
      res

let dump_local_env oc e =
   let pt i t = Printf.fprintf oc "#    type %s = %s\n" i (string_of_type t) in
   Hashtbl.iter pt e.lenv_types;
   let pc i t = Printf.fprintf oc "#    const %s = %s\n" i (string_of_const t) in
   Hashtbl.iter pc e.lenv_const;
   (* let pn i (n,_,_) = Printf.fprintf oc "#    node %s = %s\n" i (string_of_node_key n) in *)
   let pn i nk = Printf.fprintf oc "#    node %s = %s\n" i (string_of_node_key nk) in
   Hashtbl.iter pn e.lenv_nodes;
