

module ItemKeyMap = struct
	include Map.Make (
		struct
			type t = Lic.item_key
			let compare = compare
		end
	)
	let dummy () = "dummy item: add things below to complete the module"
end

module NodeKeyMap = struct
	include Map.Make (
		struct
			type t = Lic.node_key
			let compare = compare
		end
	)
	let dummy () = "dummy item: add things below to complete the module"
end

type t = {
  types  : Lic.type_ ItemKeyMap.t;
  consts : Lic.const ItemKeyMap.t;
  nodes  : Lic.node_exp NodeKeyMap.t;
  (* consts : (Lic.item_key, Lic.const    Lic.check_flag) Hashtbl.t; *)
  (* nodes  : (Lic.node_key, Lic.node_exp Lic.check_flag) Hashtbl.t; *)
}

let empty = {
   types = ItemKeyMap.empty;
   consts = ItemKeyMap.empty;
   nodes = NodeKeyMap.empty
}

(* KoKêterie : pour changer ds suffixes 
   numériques, produit :
   a, b, c, d, e, aa, ab, ac, ad, ae, etc ...
Pour tester :
for i = 0 to 42 do
Printf.printf "%3d -> %s\n" i (pretty_sfx i)
done
*)
let rec pretty_sfx i =
   if i = 0 then ""
   else
      (pretty_sfx ((i-1)/5))^(Char.escaped (char_of_int (97 + (i-1) mod 5)))

(** CREER DES IDENTS TOUT FRAIS *)
let fresh_type_id this pname pfx =
   let rec fresh x =
      let id = Printf.sprintf "%s%s" pfx (pretty_sfx x) in
      let res = Ident.make_long pname id in
      if ItemKeyMap.mem res this.types then fresh (x+1)
      else res
   in
   fresh 0

(** RECHERCHE *)
let find_type  this k = ItemKeyMap.find k this.types
let find_const this k = ItemKeyMap.find k this.consts
let find_node  this k = NodeKeyMap.find k this.nodes
   
let (find_var : Ident.t -> Lic.node_exp -> Lic.var_info option) = 
  fun id ne -> 
    let name_matches vi = vi.Lic.var_name_eff = id in
    try Some (List.find name_matches ne.Lic.inlist_eff) with Not_found -> 
    try Some (List.find name_matches ne.Lic.outlist_eff) with Not_found -> 
    match ne.Lic.loclist_eff with
      | None -> None
      | Some vil -> 
        try Some (List.find name_matches vil)
        with Not_found -> None


(** PARCOURS *)
let fold_consts (f: Lic.item_key -> Lic.const -> 'a -> 'a) (this:t) (accin:'a) : 'a =
   ItemKeyMap.fold f this.consts accin
let fold_types (f: Lic.item_key -> Lic.type_ -> 'a -> 'a) (this:t) (accin:'a) : 'a =
   ItemKeyMap.fold f this.types accin
let fold_nodes (f: Lic.node_key -> Lic.node_exp -> 'a -> 'a) (this:t) (accin:'a) : 'a =
   NodeKeyMap.fold f this.nodes accin

let iter_consts (f: Lic.item_key -> Lic.const -> unit) (this:t) : unit =
   ItemKeyMap.iter f this.consts
let iter_types (f: Lic.item_key -> Lic.type_ -> unit) (this:t) : unit = 
   ItemKeyMap.iter f this.types
let iter_nodes (f: Lic.node_key -> Lic.node_exp -> unit) (this:t) : unit =
   NodeKeyMap.iter f this.nodes



let add_type (k:Lic.item_key) (v:Lic.type_) (prg:t) : t =
   { prg with types = ItemKeyMap.add k v prg.types }

let add_const (k:Lic.item_key) (v:Lic.const) (prg:t) : t =
   { prg with consts = ItemKeyMap.add k v prg.consts }

let add_node (k:Lic.node_key) (v:Lic.node_exp) (prg:t) : t =
Verbose.printf ~level:3 "## LicPrg.add_node %s\n" (LicDump.string_of_node_key_rec k);
   { prg with nodes = NodeKeyMap.add k v prg.nodes }


let dump_entete oc = 
  let time = Unix.localtime (Unix.time ()) in
  let sys_call, _ = Array.fold_left 
    (fun (acc,i) x -> 
       if 70 < i + (String.length x) then 
	 acc ^ "\n--\t\t" ^ x, String.length ("\n--\t\t" ^ x)
       else 
	 acc ^ " " ^ x , (i+1+(String.length x))
    )
    ("",0) 
    Sys.argv
   and
      date = Printf.sprintf "%d/%02d/%02d"
         (1900+time.Unix.tm_year)
         (time.Unix.tm_mon+1)
         (time.Unix.tm_mday)
   and
      time_str = Printf.sprintf "%02d:%02d:%02d"
         (time.Unix.tm_hour)
         (time.Unix.tm_min)
         (time.Unix.tm_sec) 
   (*   and user = Unix.getlogin () *)
   and hostname = Unix.gethostname ()
   in
      (* Printf.fprintf oc "-- lus2lic version %s\n" Version.str; *)
      (* Printf.fprintf oc "-- cmd: %s\n" sys_call; *)
      (* Printf.fprintf oc "-- host: %s date: %s time: %s\n" hostname date time_str  *)
      Printf.fprintf oc "-- This file was generated by lus2lic version %s.\n" Version.str;
      Printf.fprintf oc "-- %s\n" sys_call;
      Printf.fprintf oc "-- on %s the %s at %s\n" hostname date time_str 

let to_file (oc: out_channel) (this:t) = 
   dump_entete oc;
   (* On imprime dans l'ordre du iter, donc pas terrible ???
   *)
   ItemKeyMap.iter
      (fun tn te -> output_string !Global.oc (LicDump.type_decl tn te)) this.types;
   ItemKeyMap.iter
      (fun cn ce -> output_string !Global.oc (LicDump.const_decl cn ce)) this.consts;

   (* Pour les noeuds, pas sur que ça marche tant qu'on n'a
      pas séparés les transformations  source_to_source du LicTab:
      en cas d'expansion, il y avait cette remarque :
           nb: we print res_struct, but do not return it from
           node_check, because the structure and array expansion
           modify (instanciate) the node profiles.

      On n'affiche PAS les extern Lustre::...
   *)

   NodeKeyMap.iter (
      fun _ nexp ->
         match nexp.Lic.node_key_eff with
         (* inutile d'écrire les noeuds predefs *)
         | (("Lustre",_),[]) -> ()
         | _ -> output_string !Global.oc (LicDump.node_of_node_exp_eff nexp)
   )
   this.nodes

(* GENERATEUR DE NOM DE VARIABLES *)
type id_generator = string -> string

let fresh_var_id_generator : t -> Lic.node_exp -> id_generator =
  fun prg ne ->
    let cpt = ref 0 in
    let forbidden = Hashtbl.create 100 in
    let _ = iter_consts (fun i c -> match c with
      | Lic.Extern_const_eff (s,_)
      | Lic.Abstract_const_eff (s,_,_,_)
      | Lic.Enum_const_eff   (s,_) -> Hashtbl.add forbidden (snd s) ()
      | _ -> ()
    ) prg in
    let dovar vi = Hashtbl.add forbidden vi.Lic.var_name_eff () in
    let _ = List.iter dovar ne.Lic.inlist_eff in
    let _ = List.iter dovar ne.Lic.outlist_eff in
    let _ = match ne.Lic.loclist_eff with
      | Some l -> List.iter dovar l | None -> ()
    in
    let rec dogen (pfx: string) : string =
      let id = Printf.sprintf "%s%02d" pfx !cpt in
      incr cpt;
      try (
        let _ = Hashtbl.find forbidden id in
        dogen pfx
      ) with Not_found -> (
        Hashtbl.add forbidden id ();
        id 
      )
    in
    dogen
